using Osci.Exceptions;
using Osci.Helper;

namespace Osci.Cryptographic
{
    /// <summary>Diese abstrakte Klasse stellt die Schnittstelle der Bibliothek für
    /// die Anbindung von Entschlüsselungs-Modulen (Crypto-Token) dar. Anwendungen, die
    /// OSCI-Nachrichten entschlüsseln wollen, müssen an das entschlüsselnde Rollenobjekt
    /// eine Implementation dieser Klasse übergeben.
    /// 
    /// <p>Copyright © 2018 Freie Hansestadt Bremen</p>
    /// <p>Erstellt von Governikus GmbH &amp; Co. KG</p>
    /// <p>Diese Bibliothek kann von jedermann nach Maßgabe der Bremer Lizenz für
    /// freie Softwarebibliotheken genutzt werden.</p><p>Die Lizenzbestimmungen können
    /// unter der URL www.osci.de abgerufen oder bei der
    /// Koordinierungsstelle für IT-Standards (KoSIT), Schillerstr. 1,
    /// 28195 Bremen schriftlich angefordert werden.</p>
    /// 
    /// <p>Author: P. Ricklefs, N. Büngener</p> 
    /// <p>Version: 1.8.3</p>
    /// </summary>
    public abstract class Decrypter
    {

        /// <summary> Liefert die Versionsnummer.
        /// </summary>
        /// <value> Versionsnummer
        /// </value>
        public abstract string Version
        {
            get;
        }

        /// <summary> Liefert den Namen des Herstellers.
        /// </summary>
        /// <value> Herstellername
        /// </value>
        public abstract string Vendor
        {
            get;
        }

        /// <summary> Die Implementierung dieser Methode muss das Verschlüsselungszertifikat zurückgeben.
        /// </summary>
        /// <value> Verschlüsselungszertifikat
        /// </value>
        public abstract X509Certificate Certificate
        {
            get;
        }

        /// <summary> Die Implementierung dieser Methode muss das übergebene Byte-Array gemäß RSAES-PKCS1-v1_5 entschlüsseln.
        /// </summary>
        /// <param name="data">Zu entschlüsselndes Byte-Array.
        /// </param>
        /// <returns> Entschlüsseltes Byte-Array.
        /// </returns>
        public abstract byte[] Decrypt(byte[] data);

        /// <summary> Die Implementierung dieser Methode muss das übergebene Byte-Array gemäß RSAES-OAEP oder RSAES-PKCS1-v1_5
        /// entschlüsseln. Diese Methode muss überschrieben werden, die default-Implementierung 
        /// wirft ein UnsupportedOperationException.
        /// </summary>
        /// <param name="data">Zu entschlüsselndes Byte-Array
        /// </param>
        /// <param name="data">Identifier der Mask-Generation-Function (null für RSAES-PKCS1-v1_5)
        /// </param>
        /// <param name="data">Identifier des verwendeten Hashalgorithmus (wird ignoriert für RSAES-PKCS1-v1_5)
        /// </param>
        /// <returns> Entschlüsseltes Byte-Array.
        /// </returns>
        public virtual byte[] Decrypt(byte[] data, string mfgAlgorithm, string digestAlgorithm)
        {
            throw new UnsupportedOperationException("Entschlüsselung mit RSAES-OAEP in diesem Decrypter nicht implementiert.");
        }

    }
}